# This file was auto-generated by salt's setup on Wednesday, 14 January 2015 @ 22:01:55 UTC.

from salt.version import SaltStackVersion

__saltstack_version__ = SaltStackVersion(2014, 7, 1, 0, 0, 0, None)
