# -*- coding: utf-8 -*-
'''
An "Always Approved" eauth interface to test against, not intended for
production use
'''


def auth(username, password):
    '''
    Authenticate!
    '''
    return True
